dbc = {}

local function safeCall(F, T)
   local Success, A, B, C = pcall(F, T)
   if not Success then
      if A.message then
         error(A.message, 3)
      else
         error(A, 3)
      end
   end
   return A, B, C
end

local function checkString(Arg, Usage)
   if type(Arg) ~= 'string' then
      error('Argument should be a string\n\n'..Usage,3)
   end
end

local function checkBoolean(Arg, Usage)
   if type(Arg) ~= 'boolean'  and type(Arg) ~= 'nil' then
      error('Argument should be a boolean\n\n'..Usage,3)
   end   
end

local QueryUsage=[[
query expects one argument - a SQL string to query against the database
e.g. local R = DB:query("SELECT * FROM Patient WHERE Id = 1")
]]

local function query(P, Sql)
   checkString(Sql, QueryUsage)
   local A, B =  safeCall(db.query,{sql=Sql, api=P.api, 
      name=P.name, user=P.user, password=P.password,
      live=true})
   return A, B
end

local function checkTable(T, Usage)
   if type(T) ~= 'userdata'
    or T:nodeType() ~= 'table_collection' then
      error('1st argument must be node table tree created using db.tables\n\n'..Usage,3)
   end
end

local MergeUsage=[[
Merges table records into the database.
Returns: nothing.
Requires one parameter
    A node table tree created using db.tables
Optionally 
   A second boolean, if true, the merge will be executed in the editor

e.g. DB:merge(Out, true)
]]

local function merge(P, T, Live)
   checkTable(T, MergeUsage)
   checkBoolean(Live, MergeUsage)
   safeCall(db.merge, {data=T, api=P.api, 
      name=P.name, live=Live, user=P.user, password=P.password})
end

local ExecuteUsage=[[
Executes an ad hoc SQL statement that can alter the database.

Returns: the first result set as a data tree, if the SQL statement is a query.
   A complete collection of all result sets from the statement's queries
   is returned as a second return value.

It expects at least one argument:
   * A SQL string to query against the database
   * An optional true/false boolean argument that if true means
The query will be executed live in the editor
e.g. local R = DB:execute("DELETE FROM Patient WHERE Id = 1", true)
]]
local function execute(P, Sql, Live)
   checkString(Sql, ExecuteUsage)
   checkBoolean(Live, ExecuteUsage)   
   local A, B = safeCall(db.execute, {sql=Sql, api=P.api,
      name=P.name, user=P.user, password=P.password, live=Live})
   return A,B
end

local function checkTable(T,Usage)
   if type(T) ~= 'table' then
      error(Usage,3)
   end
end

local function checkParam(T, List, Usage)
   for i=1, #List do
      if not T[List[i]] then error('Missing parameter "'..List[i]..'"\n'..Usage, 3) end
   end
end

local Usage = [[
Creates a Connection object to interact conveniently with databases.  

Returns: A database connection object.
Accepts a table with the required entries:
   'api'      - set to the database type (e.g. db.MY_SQL or db.SQL_SERVER)
   'name'     - database name/address. For db.SQLITE, this is the database file name
   'user'     - user name (neither required nor used for db.SQLITE)
   'password' - password (neither required nor used for db.SQLITE)

e.g. local DB = dbc.Connection{api=db.MY_SQL, name='test@localhost', user='root', password='secret'}
     local R = DB:query("SELECT * FROM PATIENT WHERE ID = 1")
     DB:execute("DELETE FROM PATIENT WHERE ID = 2", true)
     DB:merge(T,true)
   
The methods query, execute and merge all take an optional third parameter which if set to true
mean that the operation should be executed 'live' within the editor.
]]

function dbc.Connection(Param)
   checkTable(Param, Usage)
   checkParam(Param, {'name','user', 'password', 'api'}, Usage)
   Param.query = query
   Param.merge = merge
   Param.execute = execute
   return Param
end